/* 	$Id: xsort.c,v 1.4 1997/07/01 08:26:13 pirx Exp $	 */

/*
 * Author: Achim Bangert (abangert@ix.urz.uni-heidelberg.de)
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#ifdef XAW3D
#include <X11/Xaw3d/MenuButton.h>
#include <X11/Xaw3d/SimpleMenu.h>
#include <X11/Xaw3d/Sme.h>
#else
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/Sme.h>
#endif
#include "Listsel.h"
#include "SmeToggle.h"

#include "global.h"
#include "xtop.h"
#include "xsort.h"

static int sort_by_pid(int *, int *);
/* sort processes by pid */

static int sort_by_size(int *, int *);
/* sort processes by size */

static int sort_by_pcpu(int *, int *);
/* sort processes by % cpu time since last refresh */

static int sort_by_time(int *, int *);
/* sort processes by cpu time */

static int sort_by_ctime(int *, int *);
/* sort processes by cpu time of process + children */

static int sort_by_user(int *, int *);
/* sort processes by user name */

static int sort_by_command(int *, int *);
/* sort processes by basename of executable */

#define NUM_SORT 7

typedef struct sort_info_s {
  field_id_t fid[NUM_SORT];
  XlswComparisonFunction cmp_fun[NUM_SORT];
} sort_info_t, *sort_info_ptr;

sort_info_t sortinfo = {
  {P_PID, P_TIME, P_CTIME, P_SIZE, P_USER, P_PCPU, P_COMMAND},
  {sort_by_pid, sort_by_time, sort_by_ctime, sort_by_size, 
   sort_by_user, sort_by_pcpu, sort_by_command}
};

Widget sort_button;
Widget sort_shell;
Widget sort_option[NUM_SORT];

static int sort_by_pid(int *p1, int *p2)
/* sort processes by pid */
{
  if(procinfo.table[*p1]->pid < procinfo.table[*p2]->pid) 
    return -1;
  if(procinfo.table[*p1]->pid > procinfo.table[*p2]->pid)
    return 1;
  return 0;
  /* never reached */
}

static int sort_by_size(int *p1, int *p2)
/* sort processes by size */
{
  if(procinfo.table[*p1]->size > procinfo.table[*p2]->size) 
    return -1;
  if(procinfo.table[*p1]->size < procinfo.table[*p2]->size)
    return 1;
  return sort_by_pid(p1, p2);
}

static int sort_by_pcpu(int *p1, int *p2)
/* sort processes by % cpu time since last refresh */
{
  if(procinfo.table[*p1]->pcpu > procinfo.table[*p2]->pcpu) 
    return -1;
  if(procinfo.table[*p1]->pcpu < procinfo.table[*p2]->pcpu)
    return 1;
  return sort_by_pid(p1, p2);
}

static int sort_by_time(int *p1, int *p2)
/* sort processes by cpu time */
{
  int t1 = procinfo.table[*p1]->utime + procinfo.table[*p1]->stime;
  int t2 = procinfo.table[*p2]->utime + procinfo.table[*p2]->stime;
  
  if(t1 > t2) 
    return -1;
  if(t1 < t2)
    return 1;
  return sort_by_pid(p1, p2);
}

static int sort_by_ctime(int *p1, int *p2)
/* sort processes by cpu time of process + children */
{
  int t1 = procinfo.table[*p1]->utime + procinfo.table[*p1]->stime +
    procinfo.table[*p1]->cutime + procinfo.table[*p1]->cstime;
  int t2 = procinfo.table[*p2]->utime + procinfo.table[*p2]->stime +
    procinfo.table[*p2]->cutime + procinfo.table[*p2]->cstime;

  if(t1 > t2) 
    return -1;
  if(t1 < t2)
    return 1;
  return sort_by_pid(p1, p2);
}

static int sort_by_user(int *p1, int *p2)
/* sort processes by user name */
{
  int ret = strcmp(procinfo.table[*p1]->user, procinfo.table[*p2]->user);
  
  if(ret)
    return ret;
  else
    return sort_by_pid(p1, p2);
}

static int sort_by_command(int *p1, int *p2)
/* sort processes by basename of executable */
{
  int ret = strcmp(procinfo.table[*p1]->cmd, procinfo.table[*p2]->cmd);
  
  if(ret)
    return ret;
  else
    return sort_by_pid(p1, p2);
}

static void sort_callback(Widget w, XtPointer client, XtPointer call)
/* select field used for sorting the processes */
{
  int i = (int) client;
  int reason = (int) call;

  switch(reason)
    {
    case XstoTURNOFF:
      /* user tried to turn off sorting - forbidden; switch to default */
      XstoSetCurrent(w, (XtPointer) sort_by_pcpu);
    case XstoRADIO_TURNOFF:
      return;
    default:
      XlswSetComparisonFunction(proc_selector, sortinfo.cmp_fun[i]);
    }
}

static void create_sort_options(void)
/* create the options of the sort menu */
{
  Widget radio_group = NULL;
  Arg arglist[5];
  char buffer[32];
  int i;

  XtSetArg(arglist[0], XtNleftMargin, 20);
  XtSetArg(arglist[1], XtNplacement, XtJustifyLeft);
  XtSetArg(arglist[2], XtNonBitmap, tick_bitmap);
  for(i = 0; i < NUM_SORT; ++i)
    {
      XtSetArg(arglist[3], XtNradioData, (XtPointer) sortinfo.cmp_fun[i]);
      XtSetArg(arglist[4], XtNradioGroup, radio_group);
      sprintf(buffer, "sortBy%s", field_names[sortinfo.fid[i]]);
      sort_option[i] = 
	XtCreateManagedWidget(buffer, smeToggleObjectClass, sort_shell,
			      arglist, 5);
      XtAddCallback(sort_option[i], XtNcallback, sort_callback, (XtPointer) i);
      radio_group = sort_option[0];
    }
  if(XstoGetCurrent(radio_group) == NULL)
    XstoSetCurrent(radio_group, (XtPointer) sort_by_pcpu);
}

void create_sort_menu(void)
/* create menu for selecting sort function */
{
  sort_button =
    XtCreateManagedWidget("sortButton", menuButtonWidgetClass,
			  button_bar, NULL, 0);
  sort_shell = 
    XtCreatePopupShell("sortShell", simpleMenuWidgetClass, sort_button,
		       NULL, 0);
  XtVaSetValues(sort_button, XtNmenuName, "sortShell", NULL);
  create_sort_options();
}







