/* 	$Id: xtop.c,v 1.9 1997/07/01 08:24:04 pirx Exp $	 */

/*
 * Author: Achim Bangert (abangert@ix.urz.uni-heidelberg.de)
 *
 * xtop is based on the top program of the procps-1.01 package. Though 
 * actually not much of the original code made it into the present version
 * of xtop the general design and functionality is closely related to it.
 */

/*
 * top.c              - show top CPU processes
 *
 * Copyright (c) 1992 Branko Lankester
 * Copyright (c) 1992 Roger Binns
 *
 * Snarfed and HEAVILY modified for the YAPPS (yet another /proc ps)
 * by Michael K. Johnson, johnsonm@sunsite.unc.edu.  What is used is what
 * is required to have a common interface.
 *
 * Modified Michael K Johnson's ps to make it a top program.
 * Also borrowed elements of Roger Binns kmem based top program.
 * Changes made by Robert J. Nation (nation@rocket.sanders.lockheed.com)
 * 1/93
 *
 * Modified by Michael K. Johnson to be more efficient in cpu use
 * 2/21/93
 *
 * Changed top line to use uptime for the load average.  Also
 * added SIGTSTP handling.  J. Cowley, 19 Mar 1993.
 *
 * Modified quite a bit by Michael Shields (mjshield@nyx.cs.du.edu)
 * 1994/04/02.  Secure mode added.  "d" option added.  Argument parsing
 * improved.  Switched order of tick display to user, system, nice, idle,
 * because it makes more sense that way.  Style regularized (to K&R,
 * more or less).  Cleaned up much throughout.  Added cumulative mode.
 * Help screen improved.
 *
 * Fixed kill buglet brought to my attention by Rob Hooft.
 * Problem was mixing of stdio and read()/write().  Added
 * getnum() to solve problem.
 * 12/30/93 Michael K. Johnson
 *
 * Added toggling output of idle processes via 'i' key.
 * 3/29/94 Gregory K. Nickonov
 *
 * Fixed buglet where rawmode wasn't getting restored.
 * Added defaults for signal to send and nice value to use.
 * 5/4/94 Jon Tombs.
 *
 * Modified 1994/04/25 Michael Shields <mjshield@nyx.cs.du.edu>
 * Merged previous changes to 0.8 into 0.95.
 * Allowed the use of symbolic names (e.g., "HUP") for signal input.
 * Rewrote getnum() into getstr(), getint(), getsig(), etc.
 * 
 * Modified 1995  Helmut Geyer <Helmut.Geyer@iwr.uni-heidelberg.de> 
 * added kmem top functionality (configurable fields)
 * configurable order of process display
 * Added options for dis/enabling uptime, statistics, and memory info.
 * fixed minor bugs for ELF systems (e.g. SIZE, RSS fields)
 *
 * Modified 1996/05/18 Helmut Geyer <Helmut.Geyer@iwr.uni-heidelberg.de>
 * Use of new interface and general cleanup. The code should be far more
 * readable than before.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <sys/param.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#ifdef XAW3D
#include <X11/Xaw3d/Form.h>
#include <X11/Xaw3d/Box.h>
#include <X11/Xaw3d/Command.h>
#else
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#endif
#include <X11/bitmaps/tick.bm>
#include "Listsel.h"
#include "CircMeter.h"

#include "ksym.h"
#include "sysinfo.h"
#include "version.h"
#include "devname.h"
#include "global.h"
#include "xfields.h"
#include "xsort.h"
#include "xfilter.h"
#include "xsignal.h"
#include "xtop.h"

process_info_t procinfo = {
  NULL, /* table */
  NULL, /* str */
  NULL, /* histlist */
  0,    /* num_allocated */
  0,    /* num_used */
  0     /* num_hist */
};

typedef struct system_info_s {
  unsigned **mem;
  unsigned **stat;
  unsigned int uptime, d_uptime;
  unsigned int cpu_user, pcpu_user;
  unsigned int cpu_system, pcpu_system;
  unsigned int cpu_nice, pcpu_nice;
  unsigned int cpu_idle, pcpu_idle;
  unsigned int mem_total;
  unsigned int pmem_used;
  unsigned int pmem_shared;
  unsigned int pmem_buffers;
  unsigned int pmem_free;
} system_info_t, *system_info_ptr;

system_info_t sysinfo;

Boolean psdb_opened = False;
/* changed when psdb was opened successfully */

static String fallback_rc[] = {
#include "app-defaults.h"
  NULL
};

AppResources app_resources;

static XtResource resources[] = {
  {"fieldsMenu", "FieldsMenu", XtRString, sizeof(String),
   XtOffsetOf(AppResources, fields_menu), XtRString, 
   "PID:x \n\
    user name:USER:x \n\
    cpu usage:PCPU:x \n\
    memory usage:PMEM:x \n\
    process status:STAT:x \n\
    command:COMMAND:x"
  },
  {"signalMenu", "SignalMenu", XtRString, sizeof(String),
   XtOffsetOf(AppResources, signal_menu), XtRString, 
   "terminate:15 \n\
    quit:3 \n\
    kill:9 \n\
    abort:6 \n\
    interrupt:2 \n\
    continue:18"
  },
  {"selectPIDTitle", "SelectPIDTitle", XtRString, sizeof(String),
   XtOffsetOf(AppResources, sel_pid_title), XtRString,  
   "select pids"
  },
  {"selectTTYTitle", "SelectTTYTitle", XtRString, sizeof(String),
   XtOffsetOf(AppResources, sel_tty_title), XtRString,  
   "select ttys"
  },
  {"selectUIDTitle", "SelectUIDTitle", XtRString, sizeof(String),
   XtOffsetOf(AppResources, sel_uid_title), XtRString,  
   "select uids"
  },
  {"selectSTATTitle", "SelectSTATTitle", XtRString, sizeof(String),
   XtOffsetOf(AppResources, sel_stat_title), XtRString,  
   "select stats"
  },
  {"sigBadSignalError", "SigBadSignalError", XtRString, sizeof(String),
   XtOffsetOf(AppResources, sig_signal_error), XtRString,  
   "Signal to be sent to process %d is invalid.\n"
  },
  {"sigBadProcessError", "SigBadProcessError", XtRString, sizeof(String),
   XtOffsetOf(AppResources, sig_process_error), XtRString,  
   "Process %d does not exist or is a zombie.\n"
  },
  {"sigNoPermissionError", "SigNoPermissionError", XtRString, sizeof(String),
   XtOffsetOf(AppResources, sig_permission_error), XtRString,  
   "Permission to send signal to process %d is denied.\n"
  },
  {"sigSuccess", "SigSuccess", XtRString, sizeof(String),
   XtOffsetOf(AppResources, sig_success), XtRString,  
   "Signal to process %d sent successfully.\n"
  },
  {"prioBadProcessError", "PrioBadProcessError", XtRString, sizeof(String),
   XtOffsetOf(AppResources, prio_process_error), XtRString,  
   "Process %d does not exist or is a zombie.\n"
  },
  {"prioNoPermissionError", "PrioNoPermissionError", XtRString, sizeof(String),
   XtOffsetOf(AppResources, prio_permission_error), XtRString,  
   "Permission to change priority of process %d is denied.\n"
  },
  {"prioBadAccessError", "PrioBadAccessError", XtRString, sizeof(String),
   XtOffsetOf(AppResources, prio_access_error), XtRString,  
   "Access to lower priority of process %d is restricted to superuser.\n"
  },
  {"prioSuccess", "PrioSuccess", XtRString, sizeof(String),
   XtOffsetOf(AppResources, prio_success), XtRString,  
   "Priority of process %d changed successfully.\n"
  },
  {"refreshIntervall", "RefreshIntervall", XtRInt, sizeof(int),
   XtOffsetOf(AppResources, refresh_intervall), XtRImmediate,
   (XtPointer) DEFAULT_REFRESH
  }
};

XtAppContext app_context;
Display *app_display;
Screen *app_screen;
Pixmap tick_bitmap;
XtIntervalId refresh_timer;

Widget app_shell;
Widget manager;
Widget button_bar;
Widget quit_button;
Widget cpu_meter;
Widget mem_meter;
Widget proc_selector;

field_id_t string2field(String name)
/* convert field name to array index */
{
  field_id_t i;

  for(i = 0; i < P_END; ++i)
    if(!strcmp(name, field_names[i]))
      break;
  return i;
}
       
Boolean CvtStringToField(Display *display, XrmValue *args, Cardinal *num_args,
			 XrmValue *from, XrmValue *to, XtPointer *data)
/* convert resource representation of field to array index */
{
  String name = (String) from->addr;
  static field_id_t result;

  if((result = string2field(name)) == P_END)
    {
      XtDisplayStringConversionWarning(display, name, XtRField);
      return False;
    }
  if(to->addr)
    if(to->size < sizeof(field_id_t))
      {
	to->size = sizeof(field_id_t);
	return False;
      }
    else
      *(field_id_t *) (to->addr) = result;
  else
    to->addr = (XtPointer) &result;
  to->size = sizeof(field_id_t);
  return True;
}

static void update_sysinfo(void);
/* update information about system's cpu and memory usage */

static int sort_by_pid(proc_t **p1, proc_t **p2)
/* sort by pid */
{
  if((*p1)->pid < (*p2)->pid)
    return -1;
  if((*p1)->pid > (*p2)->pid)
    return 1;
  return 0;
  /* never reached but makes gcc happy */
}

void read_proctable(process_info_ptr p, int flags, ...)
/* extract the process information stored in the proc filesystem */
{
  PROCTAB* PT = NULL;
  proc_t *tmp;
  va_list ap;
  int i = 0;

  va_start(ap, flags);
  PT = openproc_core(flags, ap);
  do 
    {
      if(p->num_allocated <= i)
	{
	  ++p->num_allocated;
	  p->table = (proc_t **)
	    XtRealloc((char *) p->table, p->num_allocated * sizeof(proc_t *));
	  p->table[i] = NULL;
	  p->str = (String *)
	    XtRealloc((char *) p->str, p->num_allocated * sizeof(String));
	  p->str[i] = NULL;
	  p->histlist = (history_list_ptr) 
	    XtRealloc((char *) p->histlist, 
		      p->num_allocated * sizeof(history_list_t));
	  p->histlist[i].pid = -1;
	}
      if(p->table[i] && p->table[i]->cmdline)
	{
	  free((char *) *(p->table[i]->cmdline));
	  p->table[i]->cmdline = NULL;
	}
      if((tmp = readproc(PT, p->table[i])))
	{
	  p->table[i] = tmp;
	  ++i;
	}
    }
  while(tmp);
  qsort(p->table, i, sizeof(proc_t *), (void *) sort_by_pid);
  if(p->table[i])
    p->table[i]->pid = -1;
  p->num_used = i;
  closeproc(PT);
}

String create_procstring(proc_t *task, int *field, int num_fields)
/* create process string containing the requested fields */
{
  char string[2048] = "";
  char tmp[1024] = "";
  char tmp2[1024] = "";
  unsigned int time;
  unsigned int pmem;
  String ret;
  int i;

  for(i = 0; i < num_fields; ++i) 
    {
      switch(field[i]) 
	{
	case P_PID:
	  sprintf(tmp, "%5d ", task->pid);
	  break;
	case P_PPID:
	  sprintf(tmp, "%5d ", task->ppid);
	  break;
	case P_UID:
	  sprintf(tmp, "%4d ", task->uid);
	  break;
	case P_USER:
	  sprintf(tmp, "%-8.8s ", task->user);
	  break;
	case P_PCPU:
	  sprintf(tmp, "%2d.%1d ", task->pcpu / 10, task->pcpu % 10);
	  break;
	case P_PMEM:
	  pmem = (task->resident * 1000) / 
	    (sysinfo.mem[meminfo_main][meminfo_total] >> 12);
	  sprintf(tmp, "%2d.%1d ", pmem / 10, pmem % 10);
	  break;
	case P_TTY:
	  sprintf(tmp, "%-3.3s ", task->ttyc);
	  break;
	case P_TTYL:
	  sprintf(tmp, "%-6.6s ", dev_to_name(task->tty));
	  break;
	case P_PRI:
	  sprintf(tmp, "%3d ", task->priority);
	  break;
	case P_NI:
	  sprintf(tmp, "%3d ", task->nice);
	  break;
	case P_SIZE:
	  sprintf(tmp, "%5d ", task->size << 2);
	  break;
	case P_SWAP:
	  sprintf(tmp, "%4d ", (task->size - task->resident) << 2);
	  break;
	case P_RSS:	/* resident not rss, it seems to be more correct */
	  sprintf(tmp, "%4d ", task->resident << 2);
	  break;
	case P_SWCHAN:
	  sprintf(tmp, "%-9.9s ", wchan(task->wchan));
	  break;
	case P_WCHAN:
	  sprintf(tmp, "%-10x", task->wchan);
	  break;
	case P_STAT:
	  sprintf(tmp, "%-4.4s ", status(task));
	  break;
	case P_TIME:
	case P_CTIME:
	  time = (task->utime + task->stime) / HZ;
	  if(field[i] == P_CTIME)
	    time += (task->cutime + task->cstime) / HZ;
	  sprintf(tmp, "%3d:%02d ", time / 60, time % 60);
	  break;
	case P_COMMANDLINE:
	  if(task->cmdline && *(task->cmdline)) 
	    {
	      int j = 0;

	      while(((task->cmdline)[j] != NULL) && (strlen(tmp2) < 1024))
		{
		  strcat(tmp2, (task->cmdline)[j]);
		  ++j; 
		}
	      sprintf(tmp, "%s", tmp2);;
	      break;
	    } 
	case P_COMMAND:
	  sprintf(tmp, "%s", task->cmd);
	  break;
	case P_FLAGS:
	  sprintf(tmp, "%8x ", task->flags);
	  break;
	}
      strcat(string, tmp);
    }
  ret = (String) XtMalloc(strlen(string) + 1);
  strcpy(ret, string);
  return ret;
}

static void update_procstring(process_info_ptr p, int *fields, int num_fields)
/* attach strings to the process table passed in p */
{
  int i = 0;

  for(;;)
    {
      if(p->str[i])
	XtFree(p->str[i]);
      if(i >= p->num_used)
	break;
      p->str[i] = create_procstring(p->table[i], fields, num_fields);
      ++i;
    }
  p->str[i] = NULL;    
}

static void update_sysinfo(void)
/* update information about system's cpu and memory usage */
{
  system_info_ptr s = &sysinfo;
  unsigned int tmp;

#define UPDATE(now, then) {				\
    tmp = (now);					\
    now = (((now) - (then)) * 1000) / s->d_uptime;	\
    then = tmp;						\
  }

  /*
   * cpu usage 
   */
  if(!(s->stat = procstat()))
    XtAppErrorMsg(app_context, "statUnreadable", "xtopStat", "XtopError", 
		  "Processing of /proc/stat failed", NULL, NULL);
  s->pcpu_user = s->stat[procstat_cpu][procstat_cpu_user];
  s->pcpu_nice = s->stat[procstat_cpu][procstat_cpu_nice]; 
  s->pcpu_system = s->stat[procstat_cpu][procstat_cpu_system];
  s->pcpu_idle = s->stat[procstat_cpu][procstat_cpu_idle];
  s->d_uptime  = s->pcpu_user + s->pcpu_nice + s->pcpu_system + s->pcpu_idle;
  tmp = s->d_uptime;
  s->d_uptime -= s->uptime;
  s->uptime = tmp;
  UPDATE(s->pcpu_user, s->cpu_user);
  UPDATE(s->pcpu_nice, s->cpu_nice);
  UPDATE(s->pcpu_system, s->cpu_system);
  UPDATE(s->pcpu_idle, s->cpu_idle);

#undef UPDATE

  /*
   * memory usage
   */
  if(!(s->mem = meminfo()))
    XtAppErrorMsg(app_context, "meminfoUnreadable", "xtopMem", "XtopError",
		  "Processing of /proc/meminfo failed", NULL, NULL);
  s->mem_total = s->mem[meminfo_main][meminfo_total] >> 10;
  s->pmem_free = s->mem[meminfo_main][meminfo_free] >> 10;
  s->pmem_shared = s->mem[meminfo_main][meminfo_shared] >> 10;
  s->pmem_buffers = s->mem[meminfo_main][meminfo_buffers] >> 10;
  s->pmem_free = (s->pmem_free * 1000) / s->mem_total;
  s->pmem_shared = (s->pmem_shared * 1000) / s->mem_total;
  s->pmem_buffers = (s->pmem_buffers * 1000) / s->mem_total;
  s->pmem_used = 1000 - s->pmem_free - s->pmem_shared - s->pmem_buffers;
}

inline static int greater_than(int *a, int *b)
/* comparison function for integers */
{
  if(*a < *b) 
    return -1;
  if(*a > *b)
    return 1;
  return 0;
}

static void compute_pcpu(void)
/* compute percentage of cpu time per process */
{
  process_info_ptr p = &procinfo;
  int d_time;
  int i, j;

  p->histlist[p->num_hist].pid = -1;
  for(i = j = 0; i < p->num_used;)
    switch(greater_than(&(p->table[i]->pid), &(p->histlist[j].pid)))
      {
      case 1:
	if(j < p->num_hist)
	  {
	    ++j;
	    continue;
	  }
      case -1:
	p->table[i]->pcpu = 0;
	++i;
	break;
      case 0:
	p->table[i]->pcpu = p->histlist[j].time;
	++i;
	++j;
	break;
      }
  for(i = 0; i < p->num_used; ++i)
    {
      d_time = p->table[i]->utime + p->table[i]->stime;
      p->histlist[i].pid = p->table[i]->pid;
      p->histlist[i].time = d_time;
      d_time -= p->table[i]->pcpu;
      p->table[i]->pcpu = (d_time > sysinfo.d_uptime)
	? (d_time * 1000) / sysinfo.uptime
	: (d_time * 1000) / sysinfo.d_uptime;
    }
  p->num_hist = p->num_used;
}

static void update_procinfo(void)
/* re-read proc fs and fill in process table */
{
  process_info_ptr p = &procinfo;
  filter_info_ptr f = &filtinfo;
  int flags; 

  switch((flags = filter_proc_flags()) & ~PROC_ANYTTY)
    {
    case PROC_PID:
      read_proctable(p, PROC_FILL | flags, f->pid.list);
      break;
    case PROC_TTY:
      read_proctable(p, PROC_FILL | flags, f->tty.list);
      break;
    case PROC_UID:
      read_proctable(p, PROC_FILL | flags, f->uid.list, f->uid.num_used);
      break;
    case PROC_STAT:
      read_proctable(p, PROC_FILL | flags, f->stat);
      break;
    default:
      read_proctable(p, PROC_FILL | flags , NULL);
    }
  update_sysinfo();
  compute_pcpu();
  update_procstring(p, fieldinfo.shown_field, fieldinfo.num_shown);
}

void refresh_display(XtPointer client, XtIntervalId *timerid)
/* refresh sysinfo/procinfo and display them */
{
  process_info_ptr p = &procinfo;
  int *selection;
  int i, j, k, n;

  selection = XlswGetSelection(proc_selector);
  /*
   * I trust this will never be NULL which implies there was displayed 
   * at least one process once
   */
  for(n = 0; selection[n] != XlswINVALID_INDEX; ++n)
    /* get IDs of selected processes */ 
    selection[n] = p->table[selection[n]]->pid;
  qsort(selection, n, sizeof(int), (comparison_fn_t) greater_than);
  update_procinfo();
  /*
   * re-translate selected PIDs into indices again
   */
  for(i = j = k = 0;  i < p->num_used && j < n;)
    switch(greater_than(&(p->table[i]->pid), selection + j))
      {
      case -1:
	++i;
	break;
      case 1:
	++j;
	break;
      case 0:
	selection[k++] = i++;
	++j;
      }
  selection[k] = XlswINVALID_INDEX;
  XtVaSetValues(cpu_meter,
		XtNarc1, sysinfo.pcpu_user,
		XtNarc2, sysinfo.pcpu_nice,
		XtNarc3, sysinfo.pcpu_system,
		NULL);
  XtVaSetValues(mem_meter,
		XtNarc1, sysinfo.pmem_used,
		XtNarc2, sysinfo.pmem_shared,
		XtNarc3, sysinfo.pmem_buffers,
		NULL);
  XtVaSetValues(proc_selector, 
		XtNchoicesList, p->str, 
		XtNnumberChoices, p->num_used,
		XtNselectedIndex, selection,
		XtNnumberSelected, k, 
		NULL);
  refresh_timer = XtAppAddTimeOut(app_context, app_resources.refresh_intervall,
				  refresh_display, NULL); 
}

static void quit_callback(Widget w, XtPointer client, XtPointer call)
/* quit program */
{
  exit(0);
}

void create_gui(void)
/* create shell and top level widgets of the user interface */
{
  Widget child;
  Arg arglist[13];
  Cardinal n = 0;

  manager =
    XtCreateManagedWidget("manager", formWidgetClass,
			  app_shell, arglist, n);
  n = 0;
  XtSetArg(arglist[n], XtNtop, XawChainTop); n++;
  XtSetArg(arglist[n], XtNbottom, XawChainTop); n++;
  XtSetArg(arglist[n], XtNleft, XawChainLeft); n++;
  XtSetArg(arglist[n], XtNright, XawChainLeft); n++;
  XtSetArg(arglist[n], XtNorientation, XtorientHorizontal); n++;
  assert(n <= XtNumber(arglist));
  button_bar =
    XtCreateManagedWidget("buttonBar", boxWidgetClass,
			  manager, arglist, n);
  n = 0;
  quit_button = 
    XtCreateManagedWidget("quitButton", commandWidgetClass,
			  button_bar, arglist, n);
  XtAddCallback(quit_button, XtNcallback, quit_callback, NULL);
  n = 0;
  XtSetArg(arglist[n], XtNtop, XawChainTop); n++;
  XtSetArg(arglist[n], XtNbottom, XawChainTop); n++;
  XtSetArg(arglist[n], XtNleft, XawChainLeft); n++;
  XtSetArg(arglist[n], XtNright, XawChainLeft); n++;
  XtSetArg(arglist[n], XtNfromVert, button_bar); n++;
  assert(n <= XtNumber(arglist));
  cpu_meter =
    XtCreateManagedWidget("cpuMeter", circMeterWidgetClass,
			  manager, arglist, n);
  n = 0;
  XtSetArg(arglist[n], XtNtop, XawChainTop); n++;
  XtSetArg(arglist[n], XtNbottom, XawChainTop); n++;
  XtSetArg(arglist[n], XtNleft, XawChainLeft); n++;
  XtSetArg(arglist[n], XtNright, XawChainLeft); n++;
  XtSetArg(arglist[n], XtNfromHoriz, cpu_meter); n++;
  XtSetArg(arglist[n], XtNfromVert, button_bar); n++;
  assert(n <= XtNumber(arglist));
  mem_meter = XtCreateManagedWidget("memMeter", circMeterWidgetClass, 
				    manager, arglist, n);
  create_fields_menu();
  n = 0;
  XtSetArg(arglist[n], XtNtop, XawChainTop); n++;
  XtSetArg(arglist[n], XtNbottom, XawChainBottom); n++;
  XtSetArg(arglist[n], XtNleft, XawChainLeft); n++;
  XtSetArg(arglist[n], XtNright, XawChainRight); n++;
  XtSetArg(arglist[n], XtNfromVert, cpu_meter); n++;
  XtSetArg(arglist[n], XtNavailableLabel, fieldinfo.header); n++;
  XtSetArg(arglist[n], XtNselectedLabel, fieldinfo.header); n++;
  XtSetArg(arglist[n], XtNchoicesList, procinfo.str); n++;
  XtSetArg(arglist[n], XtNnumberChoices, procinfo.num_used); n++;
  XtSetArg(arglist[n], XtNmanageSet, False); n++;
  XtSetArg(arglist[n], XtNmanageClosure, False); n++;
  assert(n <= XtNumber(arglist));
  proc_selector =
    XtCreateManagedWidget("procSelector", listselWidgetClass,
			  manager, arglist, n);
  child = XtNameToWidget(proc_selector, "*availableLabel");
  XtVaSetValues(child, XtNjustify, XtJustifyLeft, NULL);
  child = XtNameToWidget(proc_selector, "*selectedLabel");
  XtVaSetValues(child, XtNjustify, XtJustifyLeft, NULL);
  child = XtNameToWidget(proc_selector, "*availableViewport");
  XtVaSetValues(child, XtNuseRight, True, NULL);
  child = XtNameToWidget(proc_selector, "*selectedViewport");
  XtVaSetValues(child, XtNuseRight, True, NULL);
  create_sort_menu();
  create_filter_menu();
  create_select_menu();
  create_signal_menu();
  create_priority_menu();
  create_filtsel_popup();
  create_report_popup();
}

static void clean_up(void)
/* clean up on termination */
{
  if(psdb_opened)
    /* I don't know if that's really necessary, but ... */
    close_psdb();
}

int main(int argc, char **argv)
/* mainline */
{
  app_shell = 
    XtVaOpenApplication(&app_context, "Xtop", NULL, 0, &argc, argv, 
			fallback_rc, sessionShellWidgetClass,
			XtNtitle, "xtop",
			NULL);
  app_display = XtDisplay(app_shell);
  app_screen = DefaultScreenOfDisplay(app_display);
  tick_bitmap = XCreateBitmapFromData(app_display, 
				      RootWindowOfScreen(app_screen),
				      tick_bits, tick_width, tick_height);
  XtAppSetTypeConverter(app_context, XtRString, XtRField, CvtStringToField,
			NULL, 0, XtCacheNone, NULL);
  XtGetApplicationResources(app_shell, &app_resources, resources,
			    XtNumber(resources), NULL, 0);
  atexit(clean_up);
  set_linux_version();
  if(open_psdb()) 
    XtAppWarningMsg(app_context, "psdbMissing", "xtopMain", "XtopError",
		    "psdb not available", NULL, NULL);
  else
    psdb_opened = True;
  update_procinfo();
  create_gui();
  if(500 > app_resources.refresh_intervall || 
     app_resources.refresh_intervall > 10000)
    app_resources.refresh_intervall = DEFAULT_REFRESH;
  refresh_display(NULL, &refresh_timer);
  XtRealizeWidget(app_shell);
  XtAppMainLoop(app_context);
  exit(0);
}





